var __addDisposableResource = (this && this.__addDisposableResource) || function (env, value, async) {
    if (value !== null && value !== void 0) {
        if (typeof value !== "object" && typeof value !== "function") throw new TypeError("Object expected.");
        var dispose, inner;
        if (async) {
            if (!Symbol.asyncDispose) throw new TypeError("Symbol.asyncDispose is not defined.");
            dispose = value[Symbol.asyncDispose];
        }
        if (dispose === void 0) {
            if (!Symbol.dispose) throw new TypeError("Symbol.dispose is not defined.");
            dispose = value[Symbol.dispose];
            if (async) inner = dispose;
        }
        if (typeof dispose !== "function") throw new TypeError("Object not disposable.");
        if (inner) dispose = function() { try { inner.call(this); } catch (e) { return Promise.reject(e); } };
        env.stack.push({ value: value, dispose: dispose, async: async });
    }
    else if (async) {
        env.stack.push({ async: true });
    }
    return value;
};
var __disposeResources = (this && this.__disposeResources) || (function (SuppressedError) {
    return function (env) {
        function fail(e) {
            env.error = env.hasError ? new SuppressedError(e, env.error, "An error was suppressed during disposal.") : e;
            env.hasError = true;
        }
        var r, s = 0;
        function next() {
            while (r = env.stack.pop()) {
                try {
                    if (!r.async && s === 1) return s = 0, env.stack.push(r), Promise.resolve().then(next);
                    if (r.dispose) {
                        var result = r.dispose.call(r.value);
                        if (r.async) return s |= 2, Promise.resolve(result).then(next, function(e) { fail(e); return next(); });
                    }
                    else s |= 1;
                }
                catch (e) {
                    fail(e);
                }
            }
            if (s === 1) return env.hasError ? Promise.reject(env.error) : Promise.resolve();
            if (env.hasError) throw env.error;
        }
        return next();
    };
})(typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
    var e = new Error(message);
    return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
});
import { getQuickJS, isFail, } from "quickjs-emscripten";
import { bytesToHex } from "@noble/hashes/utils";
import { sha256 } from "@noble/hashes/sha256";
import { logger } from "@libp2p/logger";
import { assert } from "@canvas-js/utils";
import { VMError } from "./error.js";
export class VM {
    static RUNTIME_MEMORY_LIMIT = 1024 * 640; // 640kb
    static getFileURI = (file) => `canvas:${bytesToHex(sha256(file))}`;
    log = logger("canvas:vm");
    #globalCache = new Map();
    #localCache = new Set();
    static async initialize(options = {}) {
        const quickJS = await getQuickJS();
        return new VM(quickJS, options);
    }
    runtime;
    context;
    constructor(quickJS, options) {
        this.runtime = quickJS.newRuntime();
        this.context = this.runtime.newContext();
        const runtimeMemoryLimit = options.runtimeMemoryLimit ?? VM.RUNTIME_MEMORY_LIMIT;
        this.runtime.setMemoryLimit(runtimeMemoryLimit);
        const log = options.log ?? ((...args) => console.log("[vm]", ...args));
        this.setGlobalValues({
            console: this.wrapObject({ log: this.wrapFunction((...args) => void log(...args)) }),
        });
    }
    /**
     * Cleans up this VM instance.
     */
    dispose() {
        for (const [path, handle] of this.#globalCache) {
            handle.dispose();
            this.#globalCache.delete(path);
        }
        for (const handle of this.#localCache) {
            handle.dispose();
            this.#localCache.delete(handle);
        }
        this.context.dispose();
        this.runtime.dispose();
    }
    setGlobalValues(values) {
        for (const [name, handle] of Object.entries(values)) {
            handle.consume((handle) => this.context.setProp(this.context.global, name, handle));
        }
    }
    execute(script, options = {}) {
        const filename = options.uri ?? VM.getFileURI(script);
        this.unwrapResult(this.context.evalCode(script, filename, { type: "global", strict: true })).dispose();
    }
    import(module, options = {}) {
        const filename = options.uri ?? VM.getFileURI(module);
        return this.unwrapResult(this.context.evalCode(module, filename, { type: "module", strict: true }));
    }
    get = (path) => {
        const cachedHandle = this.#globalCache.get(path);
        if (cachedHandle !== undefined) {
            return cachedHandle;
        }
        const elements = path.split(".");
        const prop = elements.pop();
        assert(prop !== undefined);
        const object = elements.length > 0 ? this.get(elements.join(".")) : this.context.global;
        const handle = this.context.getProp(object, prop);
        this.#globalCache.set(path, handle);
        return handle;
    };
    /**
     * Resolving promises inside QuickJS is tricky because you have to call
     * runtime.executePendingJobs() to get the promise to resolve, so you
     * can't use await syntax even though context.resolvePromise returns a
     * native Promise. This is a utility method that lets you use await.
     */
    resolvePromise = (promise) => {
        return new Promise((resolve, reject) => {
            this.context.resolvePromise(promise).then((result) => {
                if (isFail(result)) {
                    reject(result.error.consume(this.unwrapError));
                }
                else {
                    resolve(result.value);
                }
            }, reject);
            this.runtime.executePendingJobs();
        });
    };
    call = (fn, thisArg, args) => {
        const result = this.context.callFunction(typeof fn === "string" ? this.get(fn) : fn, typeof thisArg === "string" ? this.get(thisArg) : thisArg, ...args);
        return this.unwrapResult(result);
    };
    callAsync = async (fn, thisArg, args) => {
        const resultHandle = this.call(fn, thisArg, args);
        if (this.isInstanceOf(resultHandle, this.get("Promise"))) {
            return resultHandle.consume(this.resolvePromise);
        }
        else {
            return resultHandle;
        }
    };
    unwrapError = (handle) => {
        return new VMError(this.context.dump(handle));
    };
    wrapError = (err) => {
        const error = err instanceof Error ? err : new Error(String(err));
        let constructorName = "Error";
        if (error instanceof SyntaxError) {
            constructorName = "SyntaxError";
        }
        else if (error instanceof TypeError) {
            constructorName = "TypeError";
        }
        else if (error instanceof RangeError) {
            constructorName = "RangeError";
        }
        const args = [this.context.newString(error.message)];
        try {
            const errorHandle = this.call(constructorName, this.context.null, args);
            if (error.stack !== undefined) {
                this.context.setProp(errorHandle, "stack", this.context.newString(error.stack));
            }
            return errorHandle;
        }
        finally {
            args.forEach((arg) => arg.dispose());
        }
    };
    unwrapResult = (result) => {
        if (isFail(result)) {
            throw result.error.consume(this.unwrapError);
        }
        else {
            return result.value;
        }
    };
    getBoolean = (handle) => Boolean(this.context.getNumber(handle));
    getUint8Array = (handle) => Uint8Array.from(this.call("Array.from", "Array", [handle]).consume(this.context.dump));
    newUint8Array = (value) => {
        const byteArray = Array.from(value).map((byte) => this.context.newNumber(byte));
        return this.wrapArray(byteArray).consume((handle) => this.call("Uint8Array.from", "Uint8Array", [handle]));
    };
    /**
     * Wrap an array outside a QuickJS VM by one level,
     * returning a QuickJSHandle in the host environment.
     * `wrapArray` disposes all of its handle elements.
     */
    wrapArray = (array) => {
        const arrayHandle = this.context.newArray();
        for (const elementHandle of array) {
            this.call("Array.prototype.push", arrayHandle, [elementHandle]).dispose();
            elementHandle.dispose();
        }
        return arrayHandle;
    };
    /**
     * Unwrap an array inside a QuickJS VM by one level,
     * returning a QuickJSHandle[] in the host environment.
     * `unwrapArray` does NOT dispose of the original handle.
     */
    unwrapArray = (handle, map) => {
        const length = this.context.getLength(handle) ?? 0;
        const array = new Array(length);
        for (let index = 0; index < length; index++) {
            const elementHandle = this.context
                .newNumber(index)
                .consume((indexHandle) => this.context.getProp(handle, indexHandle));
            if (map === undefined) {
                array[index] = elementHandle;
            }
            else {
                array[index] = map(elementHandle);
            }
        }
        return array;
    };
    /**
     * Wrap an object outside a QuickJS VM by one level,
     * returning a QuickJSHandle in the host environment.
     * `wrapObject` disposes all of its handle values.
     */
    wrapObject = (object) => {
        const objectHandle = this.context.newObject();
        const entries = Array.isArray(object) ? object : Object.entries(object);
        for (const [key, valueHandle] of entries) {
            valueHandle.consume((handle) => this.context.setProp(objectHandle, key, handle));
        }
        return objectHandle;
    };
    /**
     * Unwrap an object inside a QuickJS VM by one level,
     * returning a Record<string, QuickJSHandle> in the host environment.
     * `unwrapObject` does NOT dispose of the original handle.
     */
    unwrapObject = (handle, map) => {
        const object = {};
        const keys = this.call("Object.keys", this.context.null, [handle]).consume(this.unwrapArray);
        for (const keyHandle of keys) {
            const valueHandle = this.context.getProp(handle, keyHandle);
            const key = keyHandle.consume(this.context.getString);
            if (map === undefined) {
                object[key] = valueHandle;
            }
            else {
                object[key] = map(valueHandle);
            }
        }
        return object;
    };
    wrapValue = (value) => {
        if (value === undefined) {
            return this.context.undefined;
        }
        else if (value === null) {
            return this.context.null;
        }
        else if (typeof value === "boolean") {
            return value ? this.context.true : this.context.false;
        }
        else if (typeof value === "number") {
            return this.context.newNumber(value);
        }
        else if (typeof value === "string") {
            return this.context.newString(value);
        }
        else if (value instanceof Uint8Array) {
            return this.newUint8Array(value);
        }
        else if (Array.isArray(value)) {
            return this.wrapArray(value.map(this.wrapValue));
        }
        else {
            return this.wrapObject(Object.entries(value).map(([key, value]) => [key, this.wrapValue(value)]));
        }
    };
    is = (a, b) => this.call("Object.is", this.context.null, [a, b]).consume(this.getBoolean);
    isArray = (handle) => this.context.typeof(handle) === "object" &&
        this.call("Array.isArray", this.context.null, [handle]).consume(this.context.dump);
    isInstanceOf = (instanceHandle, classHandle) => this.context.typeof(instanceHandle) === "object" &&
        this.context
            .getProp(instanceHandle, "constructor")
            .consume((constructorHandle) => this.is(constructorHandle, classHandle));
    isUint8Array = (handle) => this.isInstanceOf(handle, this.get("Uint8Array"));
    unwrapValue = (handle) => {
        if (this.is(handle, this.context.undefined)) {
            return undefined;
        }
        else if (this.is(handle, this.context.null)) {
            return null;
        }
        switch (this.context.typeof(handle)) {
            case "boolean":
                return this.getBoolean(handle);
            case "number":
                return this.context.getNumber(handle);
            case "string":
                return this.context.getString(handle);
        }
        if (this.isUint8Array(handle)) {
            return this.getUint8Array(handle);
        }
        else if (this.isArray(handle)) {
            return this.unwrapArray(handle, (elementHandle) => elementHandle.consume(this.unwrapValue));
        }
        else {
            return this.unwrapObject(handle, (propertyHandle) => propertyHandle.consume(this.unwrapValue));
        }
    };
    wrapFunction = (fn) => {
        const wrap = (value) => (value === undefined ? this.context.undefined : this.wrapValue(value));
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const vm = this;
        return this.context.newFunction(fn.name, function (...args) {
            let result = undefined;
            vm.log("invoking function");
            try {
                result = fn.apply(this, args.map(vm.unwrapValue));
            }
            catch (err) {
                vm.log("caught error in wrapped function: %O", err);
                return vm.wrapError(err);
            }
            vm.log("got result");
            if (result instanceof Promise) {
                const deferred = vm.context.newPromise();
                result.then((value) => {
                    vm.log("resolving deferred promise");
                    const handle = wrap(value);
                    deferred.settled.then(() => handle.dispose());
                    deferred.resolve(handle);
                    // wrap(value).consume((handle) => deferred.resolve(handle))
                }, (err) => {
                    vm.log("rejecting deferred promise");
                    deferred.reject(vm.wrapError(err));
                });
                deferred.settled.finally(() => vm.runtime.executePendingJobs());
                return deferred.handle;
            }
            else {
                return wrap(result);
            }
        });
    };
    unwrapFunction = (handle, thisArg) => {
        const copy = handle.dup();
        this.#localCache.add(copy);
        let thisArgCopy = copy;
        if (thisArg !== undefined) {
            thisArgCopy = thisArg.dup();
            this.#localCache.add(thisArgCopy);
        }
        return (...args) => {
            const argHandles = args.map(this.wrapValue);
            try {
                const result = this.call(copy, thisArgCopy, argHandles);
                if (this.context.typeof(result) !== "undefined") {
                    return result.consume(this.unwrapValue);
                }
            }
            finally {
                argHandles.forEach((handle) => handle.dispose());
            }
        };
    };
    unwrapFunctionAsync = (handle, thisArg) => {
        const copy = this.cache(handle);
        const thisArgCopy = thisArg ? this.cache(thisArg) : copy;
        return async (...args) => {
            const argHandles = args.map(this.wrapValue);
            try {
                const result = await this.callAsync(copy, thisArgCopy, argHandles);
                if (this.context.typeof(result) !== "undefined") {
                    return result.consume(this.unwrapValue);
                }
            }
            finally {
                argHandles.forEach((handle) => handle.dispose());
            }
        };
    };
    cache = (handle) => {
        const copy = handle.dup();
        this.#localCache.add(copy);
        return copy;
    };
    isClass = (handle) => {
        const env_1 = { stack: [], error: void 0, hasError: false };
        try {
            if (this.context.typeof(handle) !== "function") {
                return false;
            }
            const getOwnPropertyDescriptor = this.get("Object.getOwnPropertyDescriptor");
            const constructorHandle = __addDisposableResource(env_1, this.context.getProp(handle, "constructor"), false);
            const prototypeStringHandle = __addDisposableResource(env_1, this.context.newString("prototype"), false);
            const propertyDescriptorHandle = __addDisposableResource(env_1, this.unwrapResult(this.context.callFunction(getOwnPropertyDescriptor, getOwnPropertyDescriptor, [
                constructorHandle,
                prototypeStringHandle,
            ])), false);
            if (this.is(propertyDescriptorHandle, this.context.undefined)) {
                return false;
            }
            const writableHandle = __addDisposableResource(env_1, this.context.getProp(propertyDescriptorHandle, "writable"), false);
            return this.getBoolean(writableHandle) === false;
        }
        catch (e_1) {
            env_1.error = e_1;
            env_1.hasError = true;
        }
        finally {
            __disposeResources(env_1);
        }
    };
}
